from ..serializers import TableUploadField as BaseTableUploadField

try:
    import pandas as pd
except ImportError:
    pd = None

__all__ = ("TableUploadField", "PandasTableUploadField")


class TableUploadField(BaseTableUploadField):
    handlers = {
        "csv": pd.read_csv,
        "xlsx": pd.read_excel,
        "xls": pd.read_excel,
        "xlsm": pd.read_excel,
        "xlsb": pd.read_excel,
        "odf": pd.read_excel,
        "ods": pd.read_excel,
        "odt": pd.read_excel,
        "sql": pd.read_sql,
        "json": pd.read_json,
        "xml": pd.read_xml,
    }

    def update_row(self, table_object: "pd.DataFrame", index, new_row: "pd.Series"):
        table_object.iloc[index] = new_row


PandasTableUploadField = TableUploadField



import inspect
from functools import wraps

from .loggers import logger


def subject_to_change(obj):
    warning_message = "{} is subject to interface change, use with caution".format(
        obj.__name__
    )
    setattr(obj, "_subject_to_change", True)

    if inspect.isclass(obj):
        old_init = obj.__init__

        @wraps(old_init)
        def new_init(self, *args, **kwargs):
            logger.warn(warning_message)
            old_init(self, *args, **kwargs)

        obj.__init__ = new_init
        return obj

    elif inspect.isfunction(obj):

        @wraps(obj)
        def new_func(*args, **kwargs):
            logger.warn(warning_message)
            return obj(*args, **kwargs)

        return new_func


def validate_row_with_serializer(serializer_class, context=None):
    if context is None:
        context = {}

    def validator(row, i, table):
        nonlocal serializer_class

        serializer = serializer_class(
            data=row.to_dict(),
            context={
                **context,
                "row": row,
                "index": i,
                "table_object": table,
            },
        )

        serializer.is_valid(raise_exception=True)
        return serializer.save()

    return validator



from django.conf import settings
from django.db import models


class AuditableMixin(models.Model):
    created_at = models.DateTimeField(auto_now_add=True)
    modified_at = models.DateTimeField(auto_now=True)

    class Meta:
        abstract = True


class AuditorsMixin(models.Model):
    created_by = models.ForeignKey(settings.AUTH_USER_MODEL, on_delete=models.CASCADE)
    modified_by = models.ForeignKey(settings.AUTH_USER_MODEL, on_delete=models.CASCADE)

    class Meta:
        abstract = True


class PublishableMixin(models.Model):
    is_published = models.BooleanField(default=False)
    published_at = models.DateTimeField()

    def is_draft(self):
        return not self.is_published

    class Meta:
        abstract = True


class PublisherMixin(models.Model):
    published_by = models.ForeignKey(settings.AUTH_USER_MODEL, on_delete=models.CASCADE)

    class Meta:
        abstract = True



from collections.abc import Iterable, Mapping

from rest_framework.renderers import JSONRenderer as BaseJsonRenderer


class JSONRenderer(BaseJsonRenderer):
    def render_success(self, data, accepted_media_type=None, renderer_context=None):
        return data

    def render_errors(self, data, accepted_media_type=None, renderer_context=None):
        field_errors = {}
        non_field_errors = []

        if isinstance(data, Mapping):
            field_errors = data

        elif isinstance(data, Iterable):  # not mapping
            non_field_errors = data

        return {
            "field_errors": field_errors,
            "non_field_errors": non_field_errors,
        }

    def render(self, data, accepted_media_type=None, renderer_context=None):
        if renderer_context and renderer_context.get("response", None):
            response = renderer_context["response"]
            if response.status_code >= 400:
                data = self.render_errors(data, accepted_media_type, renderer_context)
            else:
                data = self.render_success(data, accepted_media_type, renderer_context)

        return super().render(data, accepted_media_type, renderer_context)



import base64
import typing
from collections import abc
from enum import Enum
from inspect import isclass
from pathlib import Path

from django.utils.translation import gettext_lazy as _
from rest_framework import serializers

from djkit.private import errors
from djkit.private.utils import subject_to_change, validate_row_with_serializer
from djkit.utils import Obfuscator


class RecursiveSerializer(serializers.Serializer):
    """A serializer that can be used to recursively serialize an object."""

    def to_representation(self, value):
        """
        :param value: A serializer object, maybe a ListSerializer or a Serializer
        :return: the data from that serializer
        """
        if self.parent is None:
            raise serializers.ValidationError(
                errors.serializer_cant_be_used_independently(self)
            )

        if isinstance(self.parent, serializers.ListSerializer):
            return self.parent.parent.to_representation(value)

        return self.parent.to_representation(value)


class EnumSerializer(serializers.BaseSerializer):
    """A serializer for enums to encourage the use of IntegerChoices for improving database performance

    :param enum: The enum to serialize
    :type enum: instance of enum.Enum

    .. code-block:: python
        :caption: models.py

        class Human(models.Model):
            class Level(models.IntegerChoices):
                BEGINNER = 0
                INTERMEDIATE = 1
                ADVANCED = 2

            class MilitaryStatus(models.TextChoices):
                EXEMPTED = "exempted", "Exempted"
                SERVED = "served", "Served"
                POSTPONED = "postponed", "Postponed"


        name = models.CharField(max_length=128)
        level = models.IntegerField(choices=Level.choices)
        military_status = models.CharField(
            choices=MilitaryStatus.choices,
             max_length=128
         )

    .. code-block:: python
        :caption: serializers.py

        class HumanSerializer(serializers.ModelSerializer):
            level = EnumSerializer(enum=models.Human.Level)
            military_status = EnumSerializer(enum=models.Human.MilitaryStatus)

            class Meta:
                model = models.Human
                fields = [
                    'id',
                    'level',
                    'military_status',
                ]

    .. code-block:: python
        :caption: POST request example

        {
            "level": "BEGINNER",
            "MILITARY_STATUS": "EXEMPTED",
        }

    .. code-block:: python
        :caption: Response example

        {
            "id": 1,
            "level": "BEGINNER",
            "MILITARY_STATUS": "EXEMPTED"
        }

    - You don't have to deal with numbers if the enum is `django.db.models.IntegerChoices`

    - request.data expects any of BEGINNER, INTERMEDIATE or ADVANCED and it automatically translates it to numbers.

    - response.data automatically gets serialized from numbers to plain text.


    """

    def __init__(self, enum, *args, **kwargs):
        super().__init__(*args, **kwargs)
        assert issubclass(enum, Enum), errors.must_pass_enum_to_enum_serializer(enum)
        self.enum = enum

    def to_internal_value(self, data):
        """
        :param data: the word like "BEGINNER"
        :return: the database equivalent to that word in self.enum
        """
        try:
            return self.enum[data].value
        except KeyError:
            raise serializers.ValidationError(
                "non-existent key passed to EnumSerializer, available keys are [{}]".format(
                    ", ".join(self.allowed_inputs)
                )
            )

    def get_display(self, member):
        return member.value if isinstance(member.value, str) else member.name

    @property
    def allowed_inputs(self):
        return [self.get_display(member) for member in self.enum]

    def to_representation(self, code):
        """
        :param code: The database equivalent format
        :return: The word equivalent to that number
        """
        for member in self.enum:
            if member.value == code:
                return self.get_display(member)

        raise serializers.ValidationError(
            detail=f"code={code} has no corresponding value in enum {self.enum.__class__.__name__}"
        )


@subject_to_change
class IOSerializer(serializers.Serializer):
    """Input/Output serializer is used when having different formats for request/response.

    user_serializer = IOSerializer(
            input_serializer=PrimaryKeyRelatedField(),
            output_serializer=UserSerializer()
        )

    In the response, you'll find the user object, but in the request, you'll only send the PK.
    """

    def __init__(self, input_serializer, output_serializer, **kwargs):
        self.input_serializer = input_serializer
        self.output_serializer = output_serializer

        super().__init__(**kwargs, required=False)

    def to_representation(self, instance):
        return self.output_serializer.to_representation(instance)

    def to_internal_value(self, data):
        return self.input_serializer.to_internal_value(data)

    def bind(self, field_name, parent):
        super().bind(field_name, parent)
        self.input_serializer.bind(field_name, parent)
        self.output_serializer.bind(field_name, parent)

    def get_fields(self):
        request = self.context.get("request", None)

        if request is None:
            return self.output_serializer.get_fields()

        if request.method in ["POST", "PUT", "PATCH"]:
            if isinstance(self.input_serializer, serializers.Field):
                return {self.input_serializer.field_name: self.input_serializer}
            return self.input_serializer().get_fields()
        if isinstance(self.output_serializer, serializers.Field):
            return {self.output_serializer.field_name: self.output_serializer}
        return self.output_serializer().get_fields()


class TableUploadField(serializers.FileField):
    """A field for handling TabularUploads in any library like Pandas or Pola.rs.

    Usage:

    class MyTableUploadField(TableUploadField):
        handlers = {
            "csv": pd.read_csv,
        }

    And usage as expected

    class MySerializerNeedingTableUpload(serializers.Serializer):
        my_csv = MyTableUploadField(required=True)

        def create(self, validated_data):
            dataframe = validated_data['my_csv']

        def validate_my_csv_row(self, row, index, table_df):
            # process row as needed and return it, or raise!
            # if you return None, return value won't be used.
            # return None if you only want to validate but don't to modify the table

            return row
    """

    default_error_messages = {
        **serializers.FileField.default_error_messages,
        "row": _("row {row} is invalid: {message}"),
        "invalid_format": _("received an unexpected format={format}"),
        "format_handler": _("the format handler raised an error"),
    }
    handlers = {}
    handler_kwargs = {}
    row_validator = None

    def _validate_handler_kwargs(self, handler_kwargs):
        if handler_kwargs:
            assert isinstance(
                handler_kwargs, abc.Mapping
            ), errors.handler_kwargs_must_be_mapping(self, handler_kwargs)

    def __init__(
        self,
        row_validator=None,
        handler_kwargs=None,
        **kwargs,
    ):
        assert "write_only" not in kwargs, errors.serializer_write_only_by_default(self)
        assert "read_only" not in kwargs, errors.serializer_cant_pass_ready_only(self)

        self._validate_handler_kwargs(handler_kwargs)
        self.handler_kwargs = handler_kwargs or {}
        self.row_validator = row_validator

        super().__init__(**kwargs, write_only=True)

    def get_row_validator(self):
        """return a function to validate each row, the function takes 3 arguments (row, i, table_object)"""
        if isinstance(self.row_validator, str):
            method = getattr(self.parent, self.row_validator, None)

            if method is None:
                raise serializers.ValidationError(
                    "method {} doesn't exist on serializer {}".format(
                        self.row_validator,
                        self.parent.__class__.__name__,
                    )
                )

            return method

        if isclass(self.row_validator) and not issubclass(
            self.row_validator, serializers.Serializer
        ):
            raise serializers.ValidationError(
                "validate_row accepts only subclasses of rest_framework.serializers.Serializer"
            )

        if isclass(self.row_validator) and issubclass(
            self.row_validator, serializers.Serializer
        ):
            return validate_row_with_serializer(self.row_validator, self.context)

        if self.row_validator is None:
            method_name = "validate_{}_row".format(self.field_name)
            return getattr(self.parent, method_name, None)

        raise serializers.ValidationError(
            "validate_row can only be of types str, callable or a subclass of serializers.Serializer"
            " but got={}".format(type(self.row_validator).__name__)
        )

    def bind(self, field_name, parent):
        super().bind(field_name, parent)
        self.row_validator = self.get_row_validator()

    def fail(self, key, **kwargs):
        if key == "row":
            assert "row" in kwargs, errors.must_pass_row_when_key_is_row()
            assert "message" in kwargs, errors.must_pass_message_when_key_is_row()

            row = kwargs["row"]
            message = kwargs["message"]

            raise serializers.ValidationError(
                {
                    "original_exception": message,
                    "message": "Error in row",
                    "row": row,
                },
                code="TABLE_ROW_ERROR",
            )

        return super().fail(key, **kwargs)

    @property
    def allowed_upload_formats(self):
        return tuple((self.get_format_handlers().keys()))

    def is_allowed_format(self, fmt):
        return fmt in self.allowed_upload_formats

    @classmethod
    def get_format_handlers(cls) -> typing.Mapping[str, typing.Callable]:
        """return a mapping of format:handler"""
        return cls.handlers

    @classmethod
    def get_format_handler(cls, file):
        """Return a handler for the provided file's format
        :param file: file object from django request
        """
        file_name = file.name
        upload_format = Path(file_name).suffix.removeprefix(".")
        handlers = cls.get_format_handlers()

        if upload_format not in handlers and "*" not in handlers:
            raise serializers.ValidationError(
                "a handler for format={} was not defined".format(upload_format)
            )

        handler = handlers[upload_format]

        if not callable(handler):
            raise serializers.ValidationError(
                "handler defined for {} is not callable".format(
                    upload_format,
                )
            )

        return handlers[upload_format]

    def call_format_handler(self, data):
        """calls a handler with self.handler_kwargs passed to the initializer"""
        format_handler = self.get_format_handler(data)
        handler_kwargs = self.get_handler_kwargs(format_handler)

        try:
            return format_handler(data.file, **handler_kwargs)
        except Exception as e:
            self.fail("format_handler", message=str(e))

    def get_handler_kwargs(self, handler_or_format) -> abc.Mapping:
        """get the kwargs that get passed to the handler"""
        return self.handler_kwargs.get(handler_or_format, {})

    def update_row(self, table_object, index, new_row):
        """update a specific row using table_object and index"""
        table_object[index] = new_row

    def process_table(self, table_object):
        """do some logic on the table before you receive it in parent's validated_data"""

        if self.row_validator:
            for i, row in table_object.iterrows():
                try:
                    new_row = self.row_validator(row, i, table_object)
                    if new_row is not None:
                        self.update_row(table_object, i, new_row)
                except serializers.ValidationError as e:
                    raise serializers.ValidationError(
                        {
                            "field_errors": e.detail,
                            "row": i,
                        }
                    )

        return table_object

    def to_internal_value(self, data):
        data = super().to_internal_value(data)
        table_object = self.call_format_handler(data)
        return self.process_table(table_object)

    def to_representation(self, value):
        raise serializers.ValidationError(
            "{} is a write_only field".format(self.__class__.__name__)
        )


class DebugSerializer(serializers.Serializer):
    def to_representation(self, serializer):
        if not isinstance(serializer, serializers.Serializer):
            raise serializers.ValidationError(
                "{} is not a valid serializer instance".format(serializer.__name__)
            )

        serializer.is_valid(raise_exception=True)
        parent = getattr(serializer, "parent", None)

        return {
            "class": serializer.__class__.__name__,
            "field_name": getattr(serializer, "field_name", None),
            "fields": serializer.get_fields(),
            "instance": getattr(serializer, "instance", None),
            "initial_data": serializer.initial_data,
            "parent": self.__class__(instance=parent.data) if parent else None,
            "default_error_messages": serializer.default_error_messages,
            "errors": serializer.errors,
            "data": serializer.data,
            "parameters": {
                "allow_null": serializer.allow_null,
                "validators": self.validators,
                "context": serializer.context,
                "label": serializer.label,
                "source": serializer.source,
                "many": isinstance(serializer, serializers.ListSerializer),
            },
        }


class Base64Field(serializers.CharField):
    """A field that accepts Base64 input.

    .. to_representation returns the string
    .. to_internal_value returns base64 string
    """

    default_error_messages = {
        **serializers.CharField.default_error_messages,
        "invalid_b64_string": "string provided is not valid base64",
    }

    def __init__(self, **kwargs):
        self.encoding = kwargs.pop("encoding", "utf-8")
        self.reverse = kwargs.pop("reverse", True)
        super().__init__(**kwargs)

    def to_base(self, s):
        """return the base64 encoded string"""
        return base64.b64encode(s.encode(self.encoding)).decode(self.encoding)

    def from_base(self, s, raise_exception=True):
        """return the string from the base64"""
        try:
            return base64.b64decode(s)
        except ValueError:
            if raise_exception:
                self.fail("invalid_b64_string")

    def to_representation(self, value: str):
        if self.reverse:
            return self.to_base(value)

        return self.from_base(value, raise_exception=False)

    def to_internal_value(self, data: str):
        data = super().to_internal_value(data)
        if self.reverse:
            return self.from_base(data)

        return self.to_base(data)


class EncryptedField(serializers.CharField):
    """to be implemented"""


class HashedField(serializers.CharField):
    """to be implemented"""


class ObfuscatedFieldMixin:
    obfuscator = Obfuscator

    def __init__(self, **kwargs):
        self.cutoff = kwargs.pop("cutoff", 4)
        self.from_end = kwargs.pop("from_end", True)
        self.char = kwargs.pop("char", "*")
        super().__init__(**kwargs)

    def to_representation(self, value):
        value = super().to_representation(value)
        return self.obfuscator.obfuscate(
            value, cutoff=self.cutoff, from_end=self.from_end, char=self.char
        )


class ObfuscatedCharField(ObfuscatedFieldMixin, serializers.CharField):
    pass


class ObfuscatedEmailField(ObfuscatedFieldMixin, serializers.EmailField):
    def to_representation(self, value):
        return self.obfuscator.email(value)



import decimal
import numbers
from functools import total_ordering

from django.core import exceptions, validators
from django.db.models import Field

from djkit.dataclasses.misc import Money


@total_ordering
class NonDatabaseField(Field):
    """A field that's not stored in databases"""

    empty_values = validators.EMPTY_VALUES

    auto_created = False
    blank = True
    concrete = False
    editable = False
    unique = False

    is_relation = False
    remote_field = None

    one_to_one = None
    one_to_many = None
    many_to_one = None
    many_to_many = None

    prevent_kwargs = ["primary_key", "blank", "editable", "unique"]

    def __init__(self, *args, **kwargs):
        for kwarg_name in self.prevent_kwargs:
            passed_kwargs = []

            if kwargs.get(kwarg_name, None):
                passed_kwargs.append(kwarg_name)

            if passed_kwargs:
                raise exceptions.ValidationError(
                    "parameters [%s] not allowed in %s"
                    % (
                        ", ".join(passed_kwargs),
                        self.__class__.__name__,
                    )
                )

        super().__init__(*args, **kwargs, primary_key=False)

    def contribute_to_class(self, cls, name, **kwargs):
        super().contribute_to_class(cls, name, private_only=True)

    def clean(self, value, model_instance):
        return value


class MoneyField(NonDatabaseField):
    description = "A field that represents an amount of money and the currency used"

    def __init__(
        self,
        amount_field="price_amount",
        currency_field="price_currency",
        verbose_name=None,
    ):
        super(MoneyField, self).__init__(verbose_name=verbose_name)
        self.amount_field = amount_field
        self.currency_field = currency_field

    def __str__(self):
        return "%s(amount_field=%s, currency_field=%s)" % (
            self.__class__.__name__,
            self.amount_field,
            self.currency_field,
        )

    def __get__(self, instance, cls=None):
        if instance is None:
            return self

        amount = getattr(instance, self.amount_field)
        currency = getattr(instance, self.currency_field)
        if amount is not None and currency is not None:
            return Money(amount, currency)
        return self.get_default()

    def __set__(self, instance, value: Money | numbers.Number):
        default = self.get_default()
        currency = default.currency if default else None

        if isinstance(value, Money):
            amount = decimal.Decimal(value.amount)
        else:
            amount = value

        # update instance
        setattr(instance, self.amount_field, amount)
        setattr(instance, self.currency_field, currency)

    def get_default(self):
        default_currency = self.model._meta.get_field(self.currency_field).get_default()
        default_amount = self.model._meta.get_field(self.amount_field).get_default()

        if default_amount is None:
            return None

        return Money(default_amount, default_currency)



"""
all errors are implemented as functions, this may change in the future
that's why the file name is prefixed with _, private files are subject to interface change
"""


def serializer_cant_pass_ready_only(serializer):
    return "can't pass read_only to a write_only serializer {}".format(
        serializer.__class__.__name__,
    )


def serializer_write_only_by_default(serializer):
    return "serializer {} is already write_only by default".format(
        serializer.__class__.__name__,
    )


def handler_kwargs_must_be_mapping(handler, obj):
    return "{}.handler_kwargs must be an instance of `abc.Mapping`, got `{}`".format(
        handler.__class__.__name__,
        obj,
    )


def handlers_and_handler_mutually_exclusive():
    return (
        "attrs `handler` and `handlers` are mutually exclusive, please only provide one"
    )


def must_pass_row_when_key_is_row():
    return "must pass row in kwargs when key=row"


def must_pass_message_when_key_is_row():
    return "must pass a message for errors in rows"


def serializer_cant_be_used_independently(serializer):
    return "{} can't be used as a dependant serializer".format(
        serializer.__class__.__name__,
    )


def table_upload_handler_must_be_a_callable(field):
    return (
        "{}.handler must be a callable which accepts row, index & table_object".format(
            field.__class__.__name__,
        )
    )


def table_upload_handlers_must_be_mapping(field, got):
    return "{}.handlers must be of type collections.Mapping, got {}".format(
        field.__class__.__name__,
        got.__class__.__name__,
    )


def table_upload_no_format_handlers_found():
    return (
        "must add some format_handlers using your favourite library like"
        " pola.rs or pandas or you can implement your custom handler for it."
    )


def must_provide_serializer_instance(got):
    return "must provide serializer instance, got {}".format(got)


def must_pass_enum_to_enum_serializer(got):
    return "must pass enum to EnumSerializer, got {}".format(got)



from ..serializers import TableUploadField as BaseTableUploadField

try:
    import polars as pl
except ImportError:
    pl = None

__all__ = ("TableUploadField", "PolarsTableUploadField")


class TableUploadField(BaseTableUploadField):
    handlers = {
        "csv": pl.read_csv,
        "json": pl.read_json,
        "xlsx": pl.read_excel,
        "xls": pl.read_excel,
        "xlsm": pl.read_excel,
        "xltm": pl.read_excel,
        "xltx": pl.read_excel,
        "xlsb": pl.read_excel,
        "odf": pl.read_excel,
        "ods": pl.read_excel,
        "odt": pl.read_excel,
    }


# Just in case you have a different TableUploadField
PolarsTableUploadField = TableUploadField



class Obfuscator:
    @classmethod
    def obfuscate(cls, s: str, char="*", cutoff=4, from_end=True):
        if not isinstance(s, str):
            raise ValueError("must provide strings to obfuscator")

        if not isinstance(cutoff, int):
            raise ValueError("cutoff must be an integer")

        if cutoff == 0:
            return s

        if cutoff < 0:
            raise ValueError(
                "must provide a positive cutoff, negative indices are not supported"
            )

        if len(s) <= cutoff:
            return char * len(s)

        if from_end:
            # note that reverse indices need to be incremented in slicing
            index = -cutoff - 1
        else:
            index = cutoff

        after_cutoff, before_cutoff = s[index:], s[:index]
        return (
            (before_cutoff + (cutoff + 1) * char)
            if from_end
            else (char * cutoff + after_cutoff)
        )

    @classmethod
    def email(cls, e, char="*", cutoff=4, from_end=True):
        if "@" not in e:
            raise ValueError("email provided is not a valid email")

        host, domain = e.split("@")
        return cls.obfuscate(host, char, cutoff, from_end) + "@" + domain



from typing import Mapping

from rest_framework import viewsets
from rest_framework.mixins import (
    CreateModelMixin,
    DestroyModelMixin,
    ListModelMixin,
    RetrieveModelMixin,
    UpdateModelMixin,
)


class GenericViewSet(viewsets.GenericViewSet):
    """A ViewSet that adds support for per-action permissions"""

    permission_classes_retrieve = []
    permission_classes_list = []
    permission_classes_create = []
    permission_classes_update = []
    permission_classes_destroy = []

    # fallback, permissions for all actions
    permission_classes = []

    def get_action_perms_map(self) -> Mapping:
        """
        Returns a map of keys representing actions and values representing action permissions
        :return: dict
        """
        return {
            "create": self.permission_classes_create,
            "update": self.permission_classes_update,
            "partial_update": self.permission_classes_update,
            "destroy": self.permission_classes_destroy,
            "retrieve": self.permission_classes_retrieve,
            "list": self.permission_classes_list,
        }

    def get_permissions(self):
        """Tries to find the action in the return value of self.get_action_perms_map & if it's not found, it tries to get
        a variable named permission_classes_<action> & if there's none, the action has no permissions.

        Example:

        @action(detail=True)
        def eat(self, request, pk=None):
            pass

        This action has many ways of passing permissions, in the action decorator as usual, or override get_action_perms_map
        or add a variable named permission_classes_eat

        """
        action = self.action
        action_perms_map = self.get_action_perms_map()

        if action in action_perms_map:
            action_perms = action_perms_map[action]
        else:
            action_perms = self.permission_classes

        return [perm() for perm in action_perms]


class ModelViewSet(
    CreateModelMixin,
    RetrieveModelMixin,
    ListModelMixin,
    UpdateModelMixin,
    DestroyModelMixin,
    GenericViewSet,
):
    """
    A replacement for `restframework.viewsets.ModelViewSet`
    """

    pass
